--[[
	author: Zax
	content:
		globales: EMLIB_
		functions: Emlib.
		main UI frame: EmlibFrameMain
		XML UI Edit: EmlibFrameEdit
	todo:
--]]

---------------------------------------------------------------------------------------
------------------------------------- VARIABLES ---------------------------------------
Emlib = {} -- nommage espace
local addonName = ...

local addonVar = {
	------------------------ nécessaire !!!
	["name"] = addonName,
	["version"] = GetAddOnMetadata(addonName, "Version"), -- ne pas convertir en nombre ici !
	["chatColor"] = "6F94EC",
	["slashCmd"] = "emotelib",
	["minimap"] = "", -- nécessite le fichier _minimap.lua (laisser vide ou NIL si pas de bouton minimap)
	------------------------ UI
	["widthUI"] = 120, -- px
	["bkColor"] = "333333", -- fond
	["bkAlpha"] = 0.5, -- fond
	["maxVisibleLines"] = 40, -- nombre maxi de lignes visibles dans la main UI
	["vLine"] = 15, -- px : hauteur d'une ligne de la liste
	["maxLetters"] = 22, -- nbre caractères affichés dans une ligne
	["lineBkColor"] = "666666", -- fond lignes
	["mouseOverAlpha"] = 0.2, -- ligne
	["fontColor"] = {
		[0] = {0.95, 0.95, 0.9}, -- user defined
		[1] = {1,0.8,0}, -- builtin
	},
	------------------------ edit UI
	["editUIwidth"] = 650, -- px (zone edit)
	["editUIheight"] = 450, -- px (zone edit)
	["editUIalpha"] = 0.95,
	["editUImargin"] = 6, -- px
	-------- token liste
	["tokenWidth"] = 120, -- px
}

function Emlib.getLocal(key) -- pour traduction (voir localization.lua) : ne pas deplacer !!!
	if (EMLIB_LOCAL_TBL[key] == nil) then return key else return EMLIB_LOCAL_TBL[key] end
end

local frameList -- scroll frame

-------------------------------------------- saved et init
EMLIB_SAVEDPERCHAR = {}

function Emlib.initDataPerChar(onlySettingsBoolean, subSettingsValue)
	local tbl = {
		["settings"] = { -- nouveau v2
			["version"] = addonVar["version"],
			["debug"] = 0,
			["visible"] = 1,
			["expanded"] = 1,
			["lines"] = 12, -- nombre de lignes *visibles* (suivant resize user)
			["posX"] = 300,
			["posY"] = 700,
			["sortalpha"] = 1,
		},
		["emotes"] = { -- nouveau nom v2 (remplace "list")
			Emlib.getNewEmote("Boggle", "/boggle", "/boggle", 1, 1),
			Emlib.getNewEmote("Laugh", "laughs openly.", "shares a laugh with %t.", 0, 2),
			Emlib.getNewEmote("Whisper", "mumbles for himself.", "bends over to %t and whispers.", 0, 3),
			Emlib.getNewEmote("Repair needed", "will soon need to repair.", "tells %t that he will soon need to repair.", 0, 3),
		},
	}
	
	if (Emlib.isTrue(onlySettingsBoolean)) then
		if (subSettingsValue ~= nil) then return tbl["settings"][subSettingsValue] else return tbl["settings"] end
	else
		return tbl
	end
end

function Emlib.updateEmoteListv2() -- fcn spéciale pour passer de la v1.x à la v2
--[[
EMLIB_SAVEDPERCHAR = { -- OLD v1.x
	["visible"] = 1,
	["lines"] = 21,
	["posY"] = 350,
	["color"] = {["a"] = 0.7, ["r"] = 0.4, ["g"] = 0.4, ["b"] = 0.4},
	["inited"] = 1,
	["alphasort"] = 1,
	["version"] = "1.00",
	["posX"] = 1240,
	["locked"] = 1,
	["list"] = {...}
--]]
	local tblNewSettings = Emlib.initDataPerChar(true)
	local tblSettings = {}
	tblSettings["version"] = addonVar["version"]
	tblSettings["debug"] = tblNewSettings["debug"]
	tblSettings["visible"] = tblNewSettings["visible"]
	tblSettings["expanded"] = tblNewSettings["expanded"]
	tblSettings["lines"] = tblNewSettings["lines"]
	tblSettings["posX"] = tblNewSettings["posX"]
	tblSettings["posY"] = tblNewSettings["posY"]
	tblSettings["sortalpha"] = tblNewSettings["alphasort"]
	
	local tblEmotes = {}
	for i, data in ipairs(EMLIB_SAVEDPERCHAR["list"]) do
		local tbl = {}
		tbl["name"] = data.name
		tbl["builtin"] = data.builtin
		tbl["index"] = data.index
		if (data.builtin ~= 1) then
			if (data.target == 1) then
				tbl["cmd"] = ""
				tbl["cmdtarget"] = data.text
			else
				tbl["cmd"] = data.text
				tbl["cmdtarget"] = ""
			end
		else
			tbl["cmd"] = data.text
			tbl["cmdtarget"] = data.text
		end
		table.insert(tblEmotes, tbl)
	end
	
	wipe(EMLIB_SAVEDPERCHAR)
	EMLIB_SAVEDPERCHAR["settings"] = tblSettings
	EMLIB_SAVEDPERCHAR["emotes"] = tblEmotes
end

---------------------------------------------------------------------------------------
--------------------------------------- EVENTS ----------------------------------------
function Emlib.eventHandler(self,event,...)
	if (event == "PLAYER_LOGIN") then
		Emlib.loading()
	end
end

---------------------------------------------------------------------------------------
--------------------------------- CHARGEMENT & BINDINGS -------------------------------
function Emlib.loading()
	----------------------------- slash commandes
	SlashCmdList["EMOTELIBSHOW"] = Emlib.slashCmd
		SLASH_EMOTELIBSHOW1 = "/"..addonVar["slashCmd"]
		SLASH_EMOTELIBSHOW2 = "/emlib"

	----------------------------- bindings
	
	----------------------------- settings perso
	local updatedPrefs = false
	if (Emlib.isEmpty(EMLIB_SAVEDPERCHAR)) then
		EMLIB_SAVEDPERCHAR = Emlib.initDataPerChar(false)
		Emlib.testPrint("=== no lua file : INIT SETTINGS ===")
	else
		if (EMLIB_SAVEDPERCHAR["list"] ~= nil) then -- MAJ en v2
			Emlib.updateEmoteListv2()
			updatedPrefs = true
		end
		local tbl = Emlib.initDataPerChar(true) -- only settings
		if (Emlib.tableKeyDiff(tbl, EMLIB_SAVEDPERCHAR["settings"], "") ~= nil) then
			EMLIB_SAVEDPERCHAR["settings"] = Emlib.copyTable(tbl, false)
			Emlib.testPrint("=== diff saved data : INIT SETTINGS ===")
		end
	end

	----------------------------- suite
	Emlib.updateMainFrame()
	Emlib.createEditUI()
	Emlib.sortListAlpha() -- tri alpha (par nom) de la liste des emotes
	
	if (Emlib.isTrue(EMLIB_SAVEDPERCHAR["settings"]["visible"])) then
		Emlib.showHide(true, true) -- scroll update
		Emlib.expandUI(EMLIB_SAVEDPERCHAR["settings"]["expanded"])
		--Emlib.showHideEditUI(true)-- ///// dev
	else
		Emlib.showHide(false)
	end
	
	--------------- minimap (nécessite fichier _minimap.lua
	if (not Emlib.isEmpty(addonVar["minimap"])) then
		--Emlib.minimapCreate(addonVar["minimap"])
		--if (_G[addonVar["minimap"]] ~= nil) then Emlib.minimapOnOff(EMLIB_SAVEDPERCHAR["settings"]["minimap"]["visible"]) end
	end
	--------------- loaded
	local str = ""
	if (not EmlibFrameMain:IsVisible()) then str = " "..Emlib.getLocal("Enter /emotelib show to display emotes list.") end
	if (Emlib.debug()) then
		SetCVar("scriptErrors", 1)
		str = str..Emlib.getTxtColor(" ------ DEBUG ON.", "orange")
	end
	if (str ~= "") then Emlib.echoChatAppl(Emlib.getLocal("loaded")..". "..str, true, "D0DE21") end
	if (updatedPrefs) then Emlib.echoChatAppl(Emlib.getLocal("Emotes list updated."), true, "green") end
end

---------------------------------------------------------------------------------------
-------------------------------------- SLASH CMD --------------------------------------
function Emlib.slashCmd(arg, execLevel)
	Emlib.slashCmd_Lib(arg, execLevel)
end

function Emlib.slashCmd_getList(arg2) -- liste des commandes-slash (voir .lib)
	-- .slashCmd_showList(dontPrint) -- print la liste en print, ou la renvoit en chaine avec dontPrint à true (pour frame cfg)
	local tbl = {
--[[
		["macro"] = {
			{
				["cmd"] = {"f_cleanline"},
				["func"] = function() Npcm.cleanMarker(Npcm.entierParDefaut(arg2, 1)) end,
				["label"] = "INTERNAL: pour macro text btn secure",
			},
		},
--]]
		["user"] = {
			{
				["cmd"] = {"show", "hide", "on", "off"},
				["type"] = "",
				["func"] = function() Emlib.showHide() end,
				["label"] = Emlib.getLocal("show/hide the list"),
			},
			{
				["cmd"] = {"reset", "resetui"},
				["type"] = "",
				["func"] = function() Emlib.resetui() end,
				["label"] = Emlib.getLocal("reset addon interface"),
			},
			{
				["cmd"] = {"resetall"},
				["type"] = "",
				["func"] = function() Emlib.resetall() end,
				["label"] = Emlib.getLocal("reset *ALL* settings to default, including saved emotes"),
			},
			{
				["cmd"] = {"help", "aide", "?"},
				["type"] = "",
				["func"] = function() Emlib.help() end,
				["label"] = Emlib.getLocal("display available slash commands"),
			},
		},
		["dev"] = {
			{
				["cmd"] = {"settings"},
				["func"] = function() Emlib.printAddonSettings() end,
				["label"] = "display saved settings",
			},
			{
				["cmd"] = {"memory", "mem", "lines"},
				["func"] = function() Emlib.printAddonMemory() end,
				["label"] = "print saved lines",
			},
			{
				["cmd"] = {"test"},
				["type"] = "string",
				["func"] = function() Emlib.fcnTest(arg2) end,
				["label"] = "development test with 1 argument",
			},
			{
				["cmd"] = {"debug"},
				["type"] = "boolean",
				["func"] = function() Emlib.toggleDebug() end,
				["label"] = "toggle debug (persistent)",
			},
		},
		["DEFAULTCMD"] = "help", -- commande par défaut
		["DEFAULTARG"] = "help", -- argument par défaut d'une commande
	}
	return tbl
end

---------------------------------------------------------------------------------------
--------------------------------------- ACTIONS ---------------------------------------
function Emlib.help()
	--InterfaceOptionsFrame_OpenToCategory(CMDLINE_cfg_frame);
	Emlib.slashCmd_showList(false) -- pas d'UI (mettre à true pour ne pas printer la liste)
end

function Emlib.fcnTest(arg) -- $$$--%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	print(date("%y.%m.%d %H:%M:%S", GetServerTime()).."===================================")
end
	
function Emlib.printAddonSettings() -- dvlpt
	Emlib.echoChatAppl("settings", true, "grey")
	for key, value in pairs(EMLIB_SAVEDPERCHAR["settings"]) do
		if (type(value) == "table") then
			Emlib.echo("    "..key..":", "SYSTEM")
			for subkey, subvalue in pairs(value) do
				if (type(subvalue) == "table") then value = "TABLE" end
				Emlib.echo("        "..subkey..": "..subvalue, "SYSTEM")
			end
		else
			Emlib.echo("    "..tostring(key)..": "..tostring(value), "SYSTEM")
		end
	end
end

function Emlib.printAddonMemory(brutBoolean) -- dvlpt
	if (Emlib.isTrue(brutBoolean)) then
		Emlib.echoTable(EMLIB_SAVEDPERCHAR["emotes"], "emotes")
	else
		Emlib.echoChatAppl("memory", true, "grey")
		for key, value in pairs(EMLIB_SAVEDPERCHAR["emotes"]) do
			Emlib.echo("     ["..key.."]  index="..value["index"].."   builtin="..value["builtin"].."  name="..value["name"].."  cmd="..Emlib.truncStr(value["cmd"], 15).."  cmdTg="..Emlib.truncStr(value["cmdtarget"], 15), "SYSTEM")
		end
	end
end

function Emlib.toggleDebug() -- dvlpt
	EMLIB_SAVEDPERCHAR["settings"]["debug"] = 1 - tonumber(EMLIB_SAVEDPERCHAR["settings"]["debug"])
	Emlib.echoChatAppl("debug mode "..Emlib.booleanToStr(Emlib.debug())..".", false, "orange")
end

function Emlib.resetui()
	EMLIB_SAVEDPERCHAR["settings"] = Emlib.initDataPerChar(true) -- init settings only
	EmlibFrameMain:ClearAllPoints()
	EmlibFrameMain:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", EMLIB_SAVEDPERCHAR["settings"]["posX"], EMLIB_SAVEDPERCHAR["settings"]["posY"])
	
	Emlib.showHide(true)
	Emlib.expandUI(true)
end

function Emlib.resetall()
	Emlib.confirmWarning("Emlib.resetallConfirmed()")
end

function Emlib.resetallConfirmed()
	wipe(EMLIB_SAVEDPERCHAR)
	EMLIB_SAVEDPERCHAR = Emlib.initDataPerChar(false)
	Emlib.sortListAlpha()

	Emlib.resetui()
	--if (_G[addonVar["minimap"]] ~= nil) then Emlib.minimapOnOff(EMLIB_SAVEDPERCHAR["settings"]["minimap"]["visible"]) end
end

function Emlib.minimapOnOff(optionalBoolean) -- UNUSED
	local val
	local frame = _G[addonVar["minimap"]]
	if (optionalBoolean == nil) then
		----------- switch
		val = (not frame:IsVisible())
	else
		----------- given value
 		val = Emlib.isTrue(optionalBoolean, true)
	end
	
	if (val) then frame:Show() else frame:Hide() end
	EMLIB_SAVEDPERCHAR["settings"]["minimap"]["visible"] = Emlib.booleanToInt(val)
end

function Emlib.minimapIsVisible() -- renvoit true si un bouton minimap existe ET visible
	if (_G[addonVar["minimap"]] ~= nil) then
		return _G[addonVar["minimap"]]:IsVisible()
	else
		return false
	end
end

function Emlib.showHide(optionalBoolean, noChatBoolean)
	local val
	if (optionalBoolean == nil) then
		----------- switch
		val = (not EmlibFrameMain:IsVisible())
	else
		----------- given value
 		val = Emlib.isTrue(optionalBoolean, true)
	end
	
	if (val) then 
		EmlibFrameMain:Show()
		Emlib.setHeight(frameList, EMLIB_SAVEDPERCHAR["settings"]["lines"] * addonVar["vLine"])
		Emlib.scrollUpdate(frameList, EMLIB_SAVEDPERCHAR["emotes"]) -- peuplement emotes list
	else
		EmlibFrameMain:Hide()
		if (Emlib.isTrue(noChatBoolean)) then
		else
			Emlib.echoChatAppl(Emlib.getLocal("Enter /emotelib show to display emotes list."), false, "D0DE21")
		end
	end
	EMLIB_SAVEDPERCHAR["settings"]["visible"] = Emlib.booleanToInt(val)
end

function Emlib.expandUI(optionalBoolean)
	local val
	if (optionalBoolean == nil) then
		----------- switch
		val = (not frameList:IsVisible())
	else
		----------- given value
 		val = Emlib.isTrue(optionalBoolean, true)
	end
	
	if (val) then frameList:Show() else frameList:Hide() end
	EMLIB_SAVEDPERCHAR["settings"]["expanded"] = Emlib.booleanToInt(val)
end

function Emlib.isExpanded()
	return frameList:IsVisible()
end

---------------------------------------------------------------------------------------
------------------------------------ FCNS INTERNES ------------------------------------

function Emlib.saveSettings() -- UNUSED : save fait en direct	
end

function Emlib.saveFramePosition()
	EMLIB_SAVEDPERCHAR["settings"]["posX"] = floor(EmlibFrameMain:GetLeft() + 0.5)
	EMLIB_SAVEDPERCHAR["settings"]["posY"] = floor(EmlibFrameMain:GetTop() + 0.5)
	Emlib.testPrint("Emlib.saveFramePosition: posX="..EMLIB_SAVEDPERCHAR["settings"]["posX"].." posY="..EMLIB_SAVEDPERCHAR["settings"]["posY"])
end

function Emlib.saveMinimapPosition() -- UNUSED
	if (_G[addonVar["minimap"]] ~= nil) then
		--EMLIB_SAVEDPERCHAR["settings"]["minimap"]["posX"] = floor(_G[addonVar["minimap"]]:GetLeft() + 0.5)
		--EMLIB_SAVEDPERCHAR["settings"]["minimap"]["posY"] = floor(_G[addonVar["minimap"]]:GetTop() + 0.5)
	end
end

function Emlib.getNewEmote(name, cmd, cmdtarget, builtin, index) -- renvoit une table
	local tbl = {
		["name"] = name or "",
		["cmd"] = cmd or "",
		["cmdtarget"] = cmdtarget or "",
		["builtin"] = builtin or 0,
		["index"] = index or GetServerTime(),
	}
	return tbl
end

function Emlib.sortListAlpha()
	EMLIB_SAVEDPERCHAR["emotes"] = Emlib.sortTable(EMLIB_SAVEDPERCHAR["emotes"], "name", "A", "A")
end

function Emlib.getEmoteDefError(tblEmote) -- renvoit le nom du champ non valide, ou nil
	if (strtrim(tblEmote.name) == "") then
		return "name"
	else
		if (strtrim(tblEmote.cmd) == "") then
			return "cmd"
		else
			if (Emlib.isEmpty(strtrim(tblEmote.cmdtarget)) and tblEmote.builtin ~= 1) then return "cmdtarget" end
		end
	end
	return nil -- aucune erreur
end

function Emlib.confirmWarning(strScriptOnAccepted)
	strScriptOnAccepted = strScriptOnAccepted or nil
	
	local str = Emlib.getLocal("Reset all settings to default?").."\n\n"..Emlib.getTxtColor(Emlib.getLocal("All of your saved emotes will be permanently deleted!"), "red")
	Emlib.showAlert("|cFF"..addonVar["chatColor"]..addonVar["name"].." "..addonVar["version"].."|r\n\n\n"..str, strScriptOnAccepted)
end

function Emlib.showAlert(msg, strScriptOnAccepted) -- simple alert
	StaticPopupDialogs["EMLIB_ALERT"] = {
		text = "%s".."\n ",
		button1 = "Reset", -- ACCEPT,
		button2 = "No", -- CANCEL,
		OnAccept = function(self)
			--print("accepted")
			if (strScriptOnAccepted ~= nil) then RunScript(strScriptOnAccepted) end
		end,
		OnCancel = function(self)
			--print("cancelled")
		end,
		showAlert = 1, -- icon warning
		timeout = 0,
		whileDead = 1,
		hideOnEscape = 1,
		preferredIndex = 3,  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
	};
	StaticPopup_Show("EMLIB_ALERT", msg)
end

---------------------------------------------------------------------------------------
------------------------------------- LIB DEDIEE --------------------------------------
function Emlib.debug() -- raccourci : renvoit true si debug ON
	return Emlib.isTrue(EMLIB_SAVEDPERCHAR["settings"]["debug"])
end

function Emlib.getVar(nomVar, subLvl1, subLvl2) -- nécessite addonVar
	-- get valeurs de la table locale addonVar
	-- si aucun argument n'est passé, renvoit la table addonVar
	if (addonVar == nil) then return nil end
	
	if (subLvl2 ~= nil) then
		if (addonVar[nomVar][subLvl1][subLvl2] ~= nil) then return addonVar[nomVar][subLvl1][subLvl2] end
	elseif (subLvl1 ~= nil) then
		if (addonVar[nomVar][subLvl1] ~= nil) then return addonVar[nomVar][subLvl1] end
	elseif (nomVar ~= nil) then
		if (addonVar[nomVar] ~= nil) then return addonVar[nomVar] end
	else
		return addonVar
	end
end

function Emlib.setVar(nomVar, valeur) -- nécessite addonVar
	if (addonVar == nil) then return nil end
	if (nomVar ~= nil) then
		if (addonVar[nomVar] ~= nil) then addonVar[nomVar] = valeur end
	end
end

---------------------------------------------------------------------------------------
------------------------------------------ UI -----------------------------------------
function Emlib.updateMainFrame() -- modif main frame : modif main framehidden par XML
	EmlibFrameMain:SetFrameStrata("LOW")
	EmlibFrameMain:SetClampedToScreen(true)
	EmlibFrameMain:SetMovable(true)
 	EmlibFrameMain:SetUserPlaced(true)
    EmlibFrameMain:SetResizable(false)
	EmlibFrameMain:ClearAllPoints()
	EmlibFrameMain:SetSize(addonVar["widthUI"], 16) -- 16 px de haut /////
	EmlibFrameMain:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", Emlib.entierParDefaut(EMLIB_SAVEDPERCHAR["settings"]["posX"], Emlib.initDataPerChar(true, "posX")), Emlib.entierParDefaut(EMLIB_SAVEDPERCHAR["settings"]["posY"], Emlib.initDataPerChar(true, "posY")))
	
	-------------------------------------------- barre titre (EmlibMenuBar) qui recouvre le frameMain
	local menubar = CreateFrame("Button", "EmlibMenuBar", EmlibFrameMain)
	menubar:SetAllPoints("$parent") -- pas besoin de SetSize()
	----------- textures
	Emlib.setTexturedFrame(menubar, "$parent_tex", Emlib.getStuffPath("Emlib_menubar", true), nil, nil, Emlib.getVar("uiAlpha"), 0.2)
	------------- titre
	menubar.text = menubar:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormalSmall")
	Emlib.setFontStyle(menubar.text, Emlib.getStuffPath("Myriad Condensed Web.ttf", true), 11, {0.3, 0.7, 0.9}, false, 1)
	menubar.text:SetPoint("LEFT", 4, 1)
	menubar.text:SetJustifyH("LEFT")
	menubar.text:SetText(addonVar["name"])
	------------- double-click : expand/collapse
	menubar:SetScript("OnDoubleClick", function(self, mouse)
		if (mouse == "LeftButton") then Emlib.expandUI() end -- switch
	end)
	------------- drag déplacement
	menubar:SetScript("OnMouseDown", function(self, mouse)
		if (not menubar.isMoving) then
			self:GetParent():StartMoving()
			self.isMoving = true
		end
	end)
	menubar:SetScript("OnMouseUp", function(self, mouse)
		if (menubar.isMoving) then
			self:GetParent():StopMovingOrSizing()
			self.isMoving = false
			Emlib.saveFramePosition()
		end
	end)
	-------------- tooltip
    menubar:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_TOP")
		GameTooltip:AddLine(Emlib.getLocal("Double-click to collapse/expand."))
		GameTooltip:Show()
    end)
    menubar:SetScript("OnLeave", function(self)
		GameTooltip:Hide()
	end)
	
	---------------------------------- POPUP MENU (barre de menu)
	------------ popmenu : frame des articles menu
	local framePopupMenu = Emlib.ui_initializeMenubarPopup("EmlibMenuBar_PopupMenu")
	-------------- arrow popup-menu
	local popupTitleBtn, btnName = Emlib.ui_createDownArrow(menubar, 0.5) -- bouton fleche menu : appel par ToggleDropDownMenu()
	popupTitleBtn:SetPoint("RIGHT", menubar, "RIGHT", -5, 0)
	popupTitleBtn:SetScript("OnClick", function(self, mouse)
		ToggleDropDownMenu(1, nil, EmlibMenuBar_PopupMenu, popupTitleBtn, -15, 3)
	end)
	
	-------------------------------------------- frame des lignes (scroll)
	framePopupMenu = Emlib.ui_initializeScrollPopup("EmlibScroll_PopupMenu")
	Emlib.ui_createUserScrollFrame(EmlibFrameMain)
end

function Emlib.ui_createDownArrow(parent, hilitedAlpha) -- création du bouton d'appel d'un menu popup
	local hilitedAlpha = tonumber(hilitedAlpha) or 0.5
	
	local btnName = parent:GetName().."_popupBtn" -- besoin du nom en retour pour l'ancre du menu popup
	local popupBtn = CreateFrame("Button", btnName, parent)
	popupBtn:SetPoint("RIGHT", parent, "RIGHT", -14, 0)
	popupBtn:SetSize(16, 16)
	--popupBtn:SetFrameStrata("MEDIUM")
	popupBtn:SetNormalTexture(Emlib.getStuffPath("arrow_down", true))
	popupBtn:SetHighlightTexture(Emlib.getStuffPath("arrow_down", true))
	popupBtn:SetPushedTexture(Emlib.getStuffPath("arrow_down", true))
	
	return popupBtn, btnName
end

function Emlib.ui_initializeMenubarPopup(frameName) -- création POPUP MENU (menu bar)
	local frameName = frameName or "__PopUpMenu"
	local frame = CreateFrame("Frame", frameName, nil, "UIDropDownMenuTemplate")
	frame.displayMode = "MENU"
	frame.initialize = function(self, level)
		local menuContent = {}
		
		menuContent.notCheckable = true
		menuContent.padding = 10 -- /////
		menuContent.isTitle = nil
		menuContent.disabled = nil
		
		if (Emlib.isExpanded()) then menuContent.text = Emlib.getLocal("Collapse") else menuContent.text = Emlib.getLocal("Expand") end
		menuContent.func = function() Emlib.expandUI() end
		UIDropDownMenu_AddButton(menuContent)
		
		menuContent.text = Emlib.getLocal("Hide")
		menuContent.func = function() Emlib.showHide(false) end
		UIDropDownMenu_AddButton(menuContent)
		
		menuContent.text = Emlib.getLocal("New...")
		menuContent.func = function() Emlib.newEmote() end
		UIDropDownMenu_AddButton(menuContent)
	
	--[[
		if (EMLIB_SAVEDPERCHAR["settings"]["sortalpha"] == 1) then menuContent.text = Emlib.getLocal("Sort by type") else menuContent.text = Emlib.getLocal("Sort alphabetical"); end;
		menuContent.notCheckable = true;
		menuContent.hasColorSwatch = false;
		menuContent.func = function() Emlib.sortList() end;
		UIDropDownMenu_AddButton(menuContent);
	--]]
	
		menuContent.text = Emlib.getLocal("Cancel")
		menuContent.colorCode = "|cffAAAADD"
		menuContent.func = function() end
		UIDropDownMenu_AddButton(menuContent)
		
		menuContent.text = Emlib.getLocal("Reset")
		menuContent.colorCode = "|cffDC571F"
		menuContent.func = function() Emlib.resetall(true) end
		UIDropDownMenu_AddButton(menuContent)
	end
	
	return frame
end

function Emlib.newEmote()
	EmlibFrameEdit.dataLineIndex = 0 -- propriété : numéro de ligne des emotes sauvegardées
	Emlib.showHideEditUI(true) -- fichier _Edit
end

function Emlib.ui_createUserScrollFrame(parent) -- scroll liste emotes
	frameList = CreateFrame("Frame", "EmlibUserScrollFrame", parent) -- définition scroll frame
	frameList:SetPoint("TOP", parent, "BOTTOM", 0, 0)
	Emlib.setColoredFrame(frameList, addonVar["bkColor"], nil, 0, addonVar["bkAlpha"])

	local scrollProps = { -- scroll frame properties (see _scroll.lua for full settings)
		-------------------- height: at least one of these properties must be set
		["height"] = EMLIB_SAVEDPERCHAR["settings"]["lines"] * addonVar["vLine"],
		["minHeight"] = 60, -- optional: must be set if frame's height can be resized (min = 60 px)
		["maxHeight"] = addonVar["maxVisibleLines"] * addonVar["vLine"], -- optional: must be set if frame's height can be resized
		["resizeButton"] = true, -- optional: if true, draw a resizing button at the frame's bottom
		["resizeScript"] = "Emlib.scrollFrameResized", -- only needed if frame's height can be resized and some other UI elements have to be updated (NIL for no script)
		--------------------
		["width"] = addonVar["widthUI"],
		["marginH"] = 0, -- px : horizontal margin from background frame
		["marginV"] = 2, -- px : vertical margin from background frame
		["alwaysShowArrow"] = false,--true, -- if true, always show scroller up and down arrows
		-------------------- lines
		["lineHeight"] = addonVar["vLine"], -- px
		["lineSpacing"] = 1, -- px
		["cellSpacing"] = 0, -- px // horizontal spacing, if several columns in a line
		-------------------- lines font
		["font"] = "arial",--Emlib.getStuffPath("Myriad Condensed Web.ttf", true), -- "friz" | "arial" | pathToLocalFont accepted
		["fontSize"] = 11,
		["fontOutline"] = false,
		["fontColor"] = nil,
		["fontAlign"] = "LEFT", -- "LEFT" | "RIGHT" | "CENTER"
		["fontAlpha"] = 0.8,
		-------------------- lines background
		["background"] = nil,
		["border"] = 0, -- only available if ["background"] is not a texture path
		["borderColor"] = nil, -- only available if ["background"] is not a texture path
		["alpha"] = 0,
		["mouseOverLineColor"] = "333333", -- visual color effect on mouseOver all the lines (nil for no mouseOver effect)
		-------------------- scroller buttons images
		["defaultScroller"] = false, -- set to true to standard WoW scroller
		["scrollerWidth"] = 10, -- px : unused if defaultScroller set to true
		["scrollerColor"] = "716BD3", -- unused if defaultScroller set to true
		["scrollerBarAlpha"] = 0.4, -- unused if defaultScroller set to true
	}
	
	local scrollColumns = { -- columns definition and properties (see _scroll.lua for full settings)
		-- column order, do not use name !
		[1] = { -- cmd
			---------- required
			["width"] = 0, -- % value accepted. if 0 or NIL : all the remaining line's width
			---------- optional
			["type"] = "Button", -- "Frame" | "Button" | "EditBox" (NIL for default frame)
			["script"] = "Emlib.scrollFrameClickLineScript",
		},
	}
	
	Emlib.scrollCreate(frameList, scrollProps, scrollColumns, nil) -- creates structure (empty scroll lines)
end

function Emlib.scrollLineUpdate(scrollFrame, lineIndex, dataLineIndex) -- do not modify args !
	local columnFrame, dataText
	if (scrollFrame == EmlibTokenScrollFrame) then
		--------------------------------------------------- liste all available token macros
		dataText = EmlibTokenScrollFrame.allCmds[dataLineIndex]
		columnFrame = scrollFrame.line[lineIndex].column[1]
		columnFrame.text:SetText(dataText)
		
	else
		--------------------------------------------------- listes emote user
		local tblEmote = EMLIB_SAVEDPERCHAR["emotes"][dataLineIndex]
		if (tblEmote ~= nil) then -- check si resize plus grand que lignes de data
			------------------------ column 1: text and texture available
			columnFrame = scrollFrame.line[lineIndex].column[1]
			local bkColor = scrollFrame.column[1].background -- bk "normal"
			local alpha = scrollFrame.column[1].alpha or 0.5 -- alpha "normal"
			local errEmote = Emlib.getEmoteDefError(tblEmote)
			if (errEmote == nil) then
				-------------- aucune erreur
				dataText = EMLIB_SAVEDPERCHAR["emotes"][dataLineIndex]["name"]
			else
				-------------- erreur
				dataText = "[!]  "..EMLIB_SAVEDPERCHAR["emotes"][dataLineIndex]["name"]
				bkColor = {0.82,0.45,0.27} -- "D17345"
				alpha = 0.6 -- outrepasser l'alpha des propriétés du scroll
			end
			
			------------ text update
			columnFrame.text:SetText(Emlib.truncStr(dataText, Emlib.getVar("maxLetters"))) -- tronquage pour affichage
			local txtColorTbl = addonVar["fontColor"][tblEmote.builtin]
			columnFrame.text:SetTextColor(txtColorTbl[1], txtColorTbl[2], txtColorTbl[3], 1)
			------------ background color update
			columnFrame.texture:SetColorTexture(bkColor[1], bkColor[2], bkColor[3], alpha)
			columnFrame.texture:SetAlpha(alpha) -- répétition pour bug SetColorTexture() ?
		end
	end
end

---------------------------------------------------------------------------------------
------------------------------------- LISTE EMOTES-------------------------------------

function Emlib.ui_initializeScrollPopup(frameName) -- création POPUP MENU (ligne scroll emotes)
	local frameName = frameName or "___PopUpMenu2"
	local frame = CreateFrame("Frame", frameName, nil, "UIDropDownMenuTemplate")
	frame.displayMode = "MENU"
	frame.initialize = function(self, level)
		----------------------------- récupération data de la ligne cliquée
		local tblEmote = {}
		local idxDataLine = frameList.indexLineClicked -- propriété variable du scroll frame
		if (idxDataLine ~= nil) then tblEmote = EMLIB_SAVEDPERCHAR["emotes"][idxDataLine] end
		----------------------------- construction articles menu
		if (tblEmote ~= {}) then
			local menuContent = {}
			menuContent.padding = 10 -- /////
			menuContent.notCheckable = true

			menuContent.isTitle = true
			menuContent.text = tblEmote.name or "  []"
			UIDropDownMenu_AddButton(menuContent)
			
			menuContent.isTitle = nil
			menuContent.disabled = nil
			
			menuContent.colorCode = "|cff"..addonVar["chatColor"]
			menuContent.text = "    "..Emlib.getLocal("Play")
			menuContent.func = function(arg1, arg2) -- arg1 renvoit le nom du frame invisible et quelques propriétés
				Emlib.ui_menuClickInLine(tblEmote, idxDataLine, "play") --arg1.value
			end
			UIDropDownMenu_AddButton(menuContent)
			
			menuContent.colorCode = "|cffC87831"
			menuContent.text = "    "..Emlib.getLocal("Edit...")
			menuContent.func = function(arg1, arg2) -- arg1 renvoit le nom du frame invisible et quelques propriétés
				Emlib.ui_menuClickInLine(tblEmote, idxDataLine, "edit") --arg1.value
			end
			UIDropDownMenu_AddButton(menuContent)
			
			menuContent.colorCode = "|cffAAAADD";
			menuContent.text = Emlib.getLocal("Cancel")
			menuContent.func = function() end
			UIDropDownMenu_AddButton(menuContent)
		end
	end
	
	return frame
end

function Emlib.scrollFrameResized(scrollFrameName, newHeight) -- MAJ prefs après un resize de la liste
	EMLIB_SAVEDPERCHAR["settings"]["lines"] = frameList.currentLines
end

function Emlib.scrollFrameClickLineScript(scrollFrameName, clickedColumnName, mouse, lineIndex) -- do not modify args !
	-- mouse: "LeftButton" | "RightButton"
	-- if (IsModifierKeyDown()...)
	local scrollFrame = _G[scrollFrameName] -- scroll frame
	--local clickedFrame = _G[clickedColumnName] -- .text and .texture available
	local dataLineIndex, dataLine = Emlib.scrollGetDataLine(scrollFrame, lineIndex) -- scroll lib
	
	if (not Emlib.isEmpty(dataLine)) then
		frameList.indexLineClicked = dataLineIndex -- propriété frame scroll pour clic dans ligne avec menu local
		if (mouse == "RightButton") then
			ToggleDropDownMenu(1, nil, EmlibScroll_PopupMenu, clickedColumnName, 80, addonVar["vLine"] + 17)
		else
			if (IsModifierKeyDown()) then
				Emlib.editEmote(dataLine, dataLineIndex)
			else
				Emlib.playEmote(dataLine)
			end
		end
	else
		Emlib.echoChatAppl("unexpected error #3800", false, "red")
	end
end

function Emlib.ui_menuClickInLine(tblEmote, numLiData, actionKind)-- menu local liste emotes
--if (Emlib.debug()) then Emlib.echoDebug(tblEmote , "----ui_menuClickInLine()") end --/////
	local actionKind = actionKind or "play" -- défaut
	if (actionKind == "play") then
		Emlib.playEmote(tblEmote)
	else
		Emlib.editEmote(tblEmote, numLiData)
	end
end

function Emlib.playEmote(tblEmote)
--test si emote d'origine par tContains(myTable, valueToSearch) in token list ? : EmlibFrameEditZone.allCmds
	local tblEmote = tblEmote or {}
	if (tblEmote ~= {}) then
		if (Emlib.getEmoteDefError(tblEmote) ~= nil) then
			Emlib.echoChatAppl(Emlib.getLocal("Error in emote definition."), false, "red")
		else
			if (tblEmote.builtin == 1) then
				----------- emote d'origine
				Emlib.executeCmd(tblEmote.cmd, true)
			else
				----------- emote perso
				Emlib.testPrint("playing user emote ---"..tblEmote["name"].."---index="..tblEmote["index"])
				local cmd = tblEmote.cmd
				if (UnitName("target") ~= nil) then
					cmd = tblEmote.cmdtarget
					if (Emlib.isEmpty(cmd)) then cmd = tblEmote.cmd end -- secu
				end
				SendChatMessage(cmd ,"EMOTE")
			end
		end
	else
		Emlib.echoChatAppl("Bad emote definition.   #1020]", false, "red")
	end
end

--[[ OLD avec DoEmote(TOKEN)
function Emlib.playEmote(tblEmote)
--test si emote d'origine par tContains(myTable, valueToSearch) in token list ? : EmlibFrameEditZone.allCmds
	local tblEmote = tblEmote or {}
	if (tblEmote ~= {}) then
		if (Emlib.getEmoteDefError(tblEmote) ~= nil) then
			Emlib.echoChatAppl(Emlib.getLocal("Error in emote definition."), false, "red")
		else
			if (tblEmote.builtin == 1) then
				----------- emote d'origine
				local cmdStr = tblEmote.cmd
				Emlib.testPrint("---playing built'in emote ---"..tblEmote["name"].."---index="..tblEmote["index"])
				if (string.sub(cmdStr, 1, 1) == "/") then cmdStr = string.sub(cmdStr, 2) end
				DoEmote(cmdStr) -- détection auto si target -- passer "none" en 2e paramètre pour ne pas tenir compte de l'éventuelle cible
			else
				----------- emote perso
				Emlib.testPrint("playing user emote ---"..tblEmote["name"].."---index="..tblEmote["index"])
				local cmd = tblEmote.cmd
				if (UnitName("target") ~= nil) then
					cmd = tblEmote.cmdtarget
					if (Emlib.isEmpty(cmd)) then cmd = tblEmote.cmd end -- secu
				end
				SendChatMessage(cmd ,"EMOTE")
			end
		end
	else
		Emlib.echoChatAppl("Bad emote definition.   #1020]", false, "red")
	end
end
--]]

function Emlib.editEmote(tblEmote, numLiData)
	local tblEmote = tblEmote or {}
	if (tblEmote ~= {}) then
		EmlibFrameEdit.dataLineIndex = numLiData -- propriété : numéro de ligne des emotes sauvegardées
		Emlib.showHideEditUI(true) -- fichier _Edit
	else
		Emlib.echoChatAppl("Bad emote definition.   #1021]", false, "red")
	end
end
