local frameTokenList

local editName, editCmd, editCmdTarget, editBuiltin -- raccourcis EditBoxes frames

function Emlib.blur(box)
	box:HighlightText(0, 0) -- dé-select
	box:ClearFocus() -- de-focus
end

function Emlib.disableObject(object, boolean)
	if (Emlib.isTrue(boolean)) then
		object:SetAlpha(0.4)
		if (object:GetObjectType() == "EditBox") then object:Disable() end
	else
		if (object:GetObjectType() == "EditBox") then object:Enable() end
		object:SetAlpha(1)
	end
end

function Emlib.disableCmdTargetZone(boolean)
	Emlib.disableObject(EmlibFrameEditZone_Label_cmdtarget, boolean)
	Emlib.disableObject(editCmdTarget, boolean)
end

function Emlib.cleanEmoteString(str) -- nettoyage de chaine emote
	if (str == nil) then
		return ""
	else
		return string.gsub(string.gsub(strtrim(str), "^ */em +", ""), "^ */e +", "") -- imbrication car pas de OR ?
	end
end

function Emlib.showHideEditUI(optionalBoolean)
	local val
	if (optionalBoolean == nil) then
		----------- switch
		val = (not EmlibFrameEdit:IsVisible())
	else
		----------- given value
 		val = Emlib.isTrue(optionalBoolean, true)
	end
	
	if (val) then
		EmlibFrameEdit:Show()
		Emlib.fillEditData()
		editName:SetFocus()
	else
		EmlibFrameEdit:Hide()
		EmlibFrameEdit.dataLineIndex = 0 -- reset
	end
end

function Emlib.fillEditData()
	local tblEmote = Emlib.getNewEmote()
	if (Emlib.strToNumber(EmlibFrameEdit.dataLineIndex) > 0) then
		tblEmote = EMLIB_SAVEDPERCHAR["emotes"][EmlibFrameEdit.dataLineIndex]
	end
	editName:SetText(tblEmote.name)
	editCmd:SetText(tblEmote.cmd)
	editCmdTarget:SetText(tblEmote.cmdtarget)
	editBuiltin.setChecked(tblEmote.builtin == 1)
	Emlib.disableCmdTargetZone(tblEmote.builtin == 1)
end

---------------------------------------------------------------------------------------
------------------------------------ UI EDITION ---------------------------------------

function Emlib.createEditUI()
	local editWidth = Emlib.getVar("editUIwidth") -- raccourci
	local tokenWidth = Emlib.getVar("tokenWidth") -- raccourci
	local editMargin = Emlib.getVar("editUImargin") -- raccourci
	
	local frameUI = CreateFrame("Frame", "EmlibFrameEdit", UIParent)
	frameUI:SetFrameStrata("MEDIUM")
	frameUI:Hide()
	frameUI:SetClampedToScreen(true)
	frameUI:SetMovable(true)
 	frameUI:SetUserPlaced(true)
    frameUI:SetResizable(false)
	frameUI:ClearAllPoints()
	frameUI:SetWidth(editWidth) -- height et placement à la fin

	-------------------------------------------- barre de titre (EmlibFrameEditMenu)
	local menubar = CreateFrame("Button", "EmlibFrameEditMenu", frameUI)
	menubar:SetSize(frameUI:GetWidth(), 16)
	menubar:SetPoint("TOPLEFT", frameUI, "TOPLEFT", 0, 0)
	Emlib.setColoredButton(menubar, {0.37,0.43,0.83}, nil, 0, Emlib.getVar("editUIalpha") + 0.05, 0.2)
	------------- titre
	local strTitle = menubar:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormalSmall")
	Emlib.setFontStyle(strTitle, "arial", 11, {1, 0.8, 0})
	strTitle:SetPoint("CENTER", 0, 0)
	strTitle:SetText(Emlib.getVar("name").." "..Emlib.getVersion(false).." - "..Emlib.getLocal("Editing"))
	strTitle:SetJustifyH("CENTER")
	strTitle:SetTextColor(0.92, 0.8, 0.1, 1)
	
	------------- drag déplacement (clic-gauche ou droit)
	menubar:EnableMouse(true)
	menubar:SetScript("OnMouseDown", function(self, mouse)
		if (not self:GetParent().isMoving) then
			self:GetParent():StartMoving()
			self:GetParent().isMoving = true
		end
	end)
	menubar:SetScript("OnMouseUp", function(self, mouse)
		if (self:GetParent().isMoving) then
			self:GetParent():StopMovingOrSizing()
			self:GetParent().isMoving = false
		end
	end)
	
	-------------------------------------------- case de fermeture
	local btnClose = CreateFrame("Button", "EmlibFrameEditBtnClose", menubar)
	btnClose:SetPoint("TOPRIGHT", menubar, "TOPRIGHT", 0, 0)
	btnClose:SetSize(30, 16)
	Emlib.setColoredButton(btnClose, {0.66,0.28,0.19}, nil, 0, Emlib.getVar("alphaUI"), 0.2)
	-------------- texte
	local strBtn = btnClose:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormalSmall")
	strBtn:SetPoint("CENTER", 1, 0)
	strBtn:SetText("X")
	strBtn:SetJustifyH("CENTER")
	------------- script
	btnClose:SetScript("OnClick", function()
		Emlib.showHideEditUI(false)
	end)
	
	-------------------------------------------- container principal (sous la menubar) : frameEditUI
	local frameEditUI = CreateFrame("Frame", "EmlibFrameEditUI", frameUI) -- zone DLOG
	frameEditUI:SetSize(editWidth, Emlib.getVar("editUIheight"))
	Emlib.setColoredFrame(frameEditUI, {0.2, 0.2, 0.2}, nil, 0, Emlib.getVar("editUIalpha"))
	frameEditUI:SetPoint("TOPLEFT", menubar, "BOTTOMLEFT", 0, 0)
	
	-------------------------------------------- zone édition
	local frame = Emlib.createEditZone(frameEditUI)
	Emlib.disableCmdTargetZone(false)
	-------------------------------------------- scroll liste token emotes
	Emlib.createTokenScroll(frameEditUI, EmlibFrameEditZone:GetHeight())
	frameTokenList:SetPoint("TOPRIGHT", frameEditUI, "TOPRIGHT", 0 - Emlib.getVar("editUImargin"), 0 - Emlib.getVar("editUImargin"))
	frameTokenList.allCmds = Emlib.getAvailableTokenEmotes() -- property
	Emlib.scrollUpdate(frameTokenList, frameTokenList.allCmds) -- peuplement liste all available token emotes
	
	-------------------------------------------- ajustements hauteurs et placement
	frameUI:SetHeight(frameEditUI:GetHeight() + menubar:GetHeight())
	frameUI:SetPoint("CENTER", 0, 0)
end

---------------------------------------------------------------------------------------
------------------------------------ ZONE EDITION -------------------------------------

function Emlib.createEditZone(parent) -- zone d'édition stricte d'une emote (inputs)
	local editMargin = Emlib.getVar("editUImargin") -- raccourci
	
	local frameEditZone = CreateFrame("Frame", "EmlibFrameEditZone", parent)
	frameEditZone:SetPoint("TOPLEFT", parent, "TOPLEFT", editMargin, 0 - editMargin)
	frameEditZone:SetSize(parent:GetWidth() - Emlib.getVar("tokenWidth") - (editMargin * 3), parent:GetHeight() - (editMargin * 2))
	Emlib.setColoredFrame(frameEditZone, "112222", nil, 0, Emlib.getVar("editUIalpha") - 0.4)
	
	-- Emlib.createEditBox(labelStr, kind, parent, labelWidth, editWidth, maxNumLines, posY)
	local posY = 20
	editName = Emlib.createEditBox(Emlib.getLocal("Name"), "name", frameEditZone, 90, 200, 1, posY)
	editName:SetScript("OnTabPressed", function(self) editCmd:SetFocus() end)
	
	posY = posY + 20 + 20
	editCmd = Emlib.createEditBox(Emlib.getLocal("Emote").."\n"..Emlib.getLocal("without target"), "cmd", frameEditZone, 90, 410, 5, posY)
	editCmd:SetScript("OnTabPressed", function(self)
		if (editCmdTarget:IsEnabled()) then editCmdTarget:SetFocus() else editName:SetFocus() end
	end)
	
	posY = posY + 70 + 20
	editCmdTarget = Emlib.createEditBox(Emlib.getLocal("Emote").."\n"..Emlib.getLocal("with target"), "cmdtarget", frameEditZone, 90, 410, 5, posY)
	editCmdTarget:SetScript("OnTabPressed", function(self) editName:SetFocus() end)

	local labelTarget = Emlib.createLabel(Emlib.getLocal("One-line emote. Use %t in emote body to point the target."), "$parent_labelTarget", frameEditZone, 200, 1, 11, {0.64,0.48,0})
	labelTarget:SetPoint("TOPLEFT", editCmdTarget, "BOTTOMLEFT", 0 , -4)

	------------------------------------------- bouton check
	local fontTbl = {["font"] = "arial", ["size"] = 12, ["color"] = Emlib.getVar("colorPerChar")}
	editBuiltin = Emlib.createCheckButton("EmlibFrameEditZone_ChkBuiltin", Emlib.getLocal("Built-in token emote"), "848442", fontTbl, frameEditZone, false)
	editBuiltin:SetPoint("LEFT", editName, "RIGHT", 50, 0)
	
	------------------------------------------- boutons
	--Emlib.createButton(btnName, parent, label, width, objAnchor, align, posX, bkColor, alpha)
	local btnMargin = 0
	local btnClose = Emlib.createButton("$parent_btnClose", frameEditZone, Emlib.getLocal("Close"), 70, editCmdTarget, "LEFT", btnMargin, {0.49,0.4,0.07}, nil)
	btnClose:SetScript("OnClick", function() Emlib.showHideEditUI(false) end)
	
	local btnSave = Emlib.createButton("$parent_btnSave", frameEditZone, Emlib.getLocal("Save"), 90, editCmdTarget, "RIGHT", btnMargin, {0.37,0.43,0.83}, nil)
	btnSave:SetScript("OnClick", function() Emlib.btnEmoteSave() end)
	btnMargin = btnMargin + btnSave:GetWidth() + 25
	
	local btnTest = Emlib.createButton("$parent_btnTest", frameEditZone, Emlib.getLocal("Test"), 70, editCmdTarget, "RIGHT", btnMargin, {0.49,0.4,0.07}, nil)
	btnTest:SetScript("OnClick", function() Emlib.btnEmoteTest() end)
	btnMargin = btnMargin + btnTest:GetWidth() + 30
	
	local btnDelete = Emlib.createButton("$parent_btnDelete", frameEditZone, Emlib.getLocal("Delete"), 90, editCmdTarget, "RIGHT", btnMargin, {0.59,0.17,0.06}, nil)
	btnDelete:SetScript("OnClick", function() Emlib.btnEmoteDelete() end)
	btnMargin = btnMargin + btnDelete:GetWidth() + 25
	
	return frameEditZone
end

function Emlib.createEditBox(labelStr, kind, parent, labelWidth, editWidth, maxNumLines, posY)
--***return parent et placement hros script ???
	-- creation d'une EditBox avec son label (renvoit l'EditBox seule)
	local function getHeight(numLines)
		local lineHeight = 12 -- /////
		return (lineHeight * numLines) + 10
	end
	
	local parentName = parent:GetName()
	local marginLabel = 6 -- /////
	------------------------- ensemble
	local fBk = CreateFrame("Frame", "$parent_Bk_"..kind, parent)
	fBk:SetSize(labelWidth + marginLabel + editWidth, getHeight(maxNumLines))
	fBk:SetPoint("TOPLEFT", parent, "TOPLEFT", 0, 0 - posY)
--Emlib.setColoredFrame(fBk, "white", nil, 0, 0.4)--///// dev
	------------------------- label
	local fLabel = CreateFrame("Frame", parentName.."_Label_"..kind, fBk)
	fLabel:SetSize(labelWidth, 16)
	fLabel:SetPoint("TOPLEFT", fBk, "TOPLEFT", 0, 0)
	------------- texte
	local str = fLabel:CreateFontString(parentName.."_"..kind.."_Txt", "OVERLAY", "GameFontNormal")
	Emlib.setFontStyle(str, "arial", 11, {0.8, 0.6, 0})
	str:SetPoint("RIGHT", 0, -3)
	str:SetJustifyH("RIGHT")
	str:SetText(labelStr)
	------------------------- editBox
	local editBox = CreateFrame("EditBox", parentName.."_EditBox_"..kind, fBk)
	editBox:SetPoint("TOPLEFT", fLabel, "TOPRIGHT", marginLabel, 0) -- SetHeight() ne fonctionne pas si multiLine !
	editBox:SetPoint("BOTTOMRIGHT", fBk, "BOTTOMRIGHT", 0,0)
	local multiLine = (maxNumLines > 1)
	editBox:SetMultiLine(multiLine)
	if (multiLine) then
		str:SetPoint("RIGHT", 0, -8)
		editBox:SetMaxLetters(255)
		editBox:SetTextInsets(6, 6, 4, 4) -- left, right, top, bottom
		editBox:SetScript("OnTextChanged", function(self)
			if (#{strsplit("\n", self:GetText())} > maxNumLines) then Emlib.blur(self) end
		end)
	else
		editBox:SetMaxLetters(30)
		editBox:SetTextInsets(6, 6, 0, 0) -- left, right, top, bottom
		editBox:SetScript("OnEnterPressed", function(self) Emlib.blur(self) end)
	end
	Emlib.setColoredFrame(editBox, "848442", "222222", 1, 0.5)
	---------------
	editBox:SetAutoFocus(false)
	editBox:SetHighlightColor(0.65,0.69,0.1) -- couleur sélection texte
	editBox:SetFontObject("ChatFontNormal")
	editBox:SetFont(Emlib.getStuffPath("consola.ttf", true), 12, "NONE")
	editBox:SetText("")
	
	editBox:SetScript("OnEscapePressed", function(self) Emlib.blur(self) end)
	------------------------- 
	return editBox
end

function Emlib.createCheckButton(btnName, label, bkColor, fontTbl, parent, checked) 
	-- btn.checked disponible (propriété)
	-- btn.setChecked(boolean) disponible
	-- btn.setLabel(labelStr) disponible
	-- btn.setLabelColor(color) disponible
	local marginH = 6
	-------------------------------------------- btn (ensemble : clic sur label = clic check box)
	local btn = CreateFrame("Button", btnName, parent)
	btn:SetSize(16, 16) -- modif width plus bas
	------------------------- case à cocher
	local checkBox = CreateFrame("Frame", btnName.."_chkBox", btn) -- un Frame pour passer le clic !
	checkBox:SetSize(16, 16)
	Emlib.setColoredFrame(checkBox, bkColor, "111111", 1, Emlib.getVar("alphaUI"), 0.2)
	checkBox:SetPoint("LEFT", btn, "LEFT", 0, 0)
	local strCheck = checkBox:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	strCheck:SetFont(Emlib.getStuffPath("arial", true), 12, "OUTLINE")
	strCheck:SetTextColor(0.86, 0.84, 0.49)
	strCheck:SetPoint("CENTER", 1, 0)
	strCheck:SetJustifyH("CENTER")
	------------------------- label
	local fLabel = CreateFrame("Button", btn:GetName().."_Label", btn)
	fLabel:SetSize(1, btn:GetHeight()) -- largeur très fine pour ne pas gêner le clic
	fLabel:SetPoint("LEFT", checkBox, "RIGHT", marginH, 0)
	------------- text
	btn.label = fLabel:CreateFontString(fLabel:GetName().."_Txt", "OVERLAY", "GameFontNormal")
	Emlib.setFontStyle(btn.label, fontTbl.font, fontTbl.size, fontTbl.color)
	btn.label:SetPoint("LEFT", 0, 0)
	btn.label:SetJustifyH("LEFT")
	btn.label:SetText(label)
	-------------------------------------------- fcns
	btn.checked = Emlib.isTrue(checked)
	btn.setChecked = function(boolean)
		btn.checked = Emlib.isTrue(boolean)
		strCheck:SetText("")
		if (btn.checked) then strCheck:SetText("X") end
	end
	btn.setLabel = function(labelStr) btn.label:SetText(labelStr) end
	btn.setLabelColor = function(color) Emlib.setFontStyle(btn.label, fontTbl.font, fontTbl.size, color) end
	-------------------------------------------- scripts
	btn:SetScript("OnClick", function(self)
		self.setChecked(not self.checked)
		Emlib.disableCmdTargetZone(self.checked) -- spécial EmoteLib
	end)
	btn:SetWidth(checkBox:GetWidth() + marginH + btn.label:GetStringWidth() + 4)
	-------------------------------------------- 
	return btn
end

function Emlib.createButton(btnName, parent, label, width, objAnchor, align, posX, bkColor, alpha) -- dessin d'un bouton standard
	local alpha = alpha or Emlib.getVar("alphaUI")
	local marginV = 60 -- par rapport au parent
	local btn = CreateFrame("Button", btnName, parent)
	if (align == "RIGHT") then posX = 0 - posX end
	btn:SetSize(width, 18)
	btn:SetPoint("TOP"..align, objAnchor, "BOTTOM"..align, posX, 0 - marginV)
	Emlib.setColoredButton(btn, bkColor, "111111", 1, alpha, 0.6)
	-------------- texte
	local strBtn = btn:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormalSmall")
	strBtn:SetFont(Emlib.getStuffPath("Myriad Condensed Web.ttf", true), 12, "NONE")
	strBtn:SetTextColor(1, 0.8, 0) -- jaune standard
	strBtn:SetPoint("CENTER", 1, 0)
	strBtn:SetText(label)
	strBtn:SetJustifyH("CENTER")
	
	return btn
end

function Emlib.createLabel(labelStr, optionalName, parent, width, numLines, fontSize, fontColor)
	-- création d'un simple frame avec une string
	-- btn.setText(labelStr) disponible
	
	local function getHeight(numLines)
		local numLines = tonumber(numLines) or 1
		local lineHeight = 12 -- /////
		return (lineHeight * numLines) + 10
	end
	
	local width = tonumber(width) or 40
	local frame = CreateFrame("Frame", optionalName, parent)
	frame:SetSize(width, getHeight(numLines))
--Emlib.setColoredFrame(frame, "yellow", nil, 0, 0.4)--///// dev
	
	------------------------ text
	local strLabel = frame:CreateFontString("$parent_Txt", "OVERLAY", "GameFontNormal")
	Emlib.setFontStyle(strLabel, "arial", fontSize, fontColor)
	strLabel:SetPoint("LEFT", 0, 0)
	strLabel:SetJustifyH("LEFT")
	strLabel:SetText(labelStr)
	------------------------ fcns
	frame.setText = function(str)
		strLabel:SetText(str)
	end
	------------------------- 
	return frame
end

function Emlib.btnEmoteTest()
	local tblEmote = Emlib.getNewEmote(editName:GetText(), editCmd:GetText(), editCmdTarget:GetText(), Emlib.booleanToInt(editBuiltin.checked), nil)
	local errEmote = Emlib.getEmoteDefError(tblEmote)
	if (errEmote == "cmdtarget") then
		tblEmote.cmdtarget = editCmd:GetText()
		editCmdTarget:SetText(tblEmote.cmdtarget)
		errEmote = nil
	end
	
	if (errEmote == nil) then
		-------------- aucune erreur
		Emlib.playEmote(tblEmote)
	else
		-------------- erreur
		Emlib.echoChatAppl(Emlib.getLocal("Error in emote definition."), false, "red")
		_G["EmlibFrameEditZone_EditBox_"..errEmote]:SetFocus()
	end
end

function Emlib.btnEmoteSave()
	local function completeEmote(tblEmote, builtinBoolean)
		if (Emlib.isTrue(builtinBoolean)) then tblEmote.cmdtarget = tblEmote.cmd end
		return tblEmote
	end
	
	local tblEmote = Emlib.getNewEmote(editName:GetText(), editCmd:GetText(), editCmdTarget:GetText(), Emlib.booleanToInt(editBuiltin.checked), nil)
	local errEmote = Emlib.getEmoteDefError(tblEmote)
	if (errEmote == "cmdtarget") then
		tblEmote.cmdtarget = editCmd:GetText()
		editCmdTarget:SetText(tblEmote.cmdtarget)
		errEmote = nil
	end
	
	local errSaved = false
	if (errEmote == nil) then
		------------------------------- aucune erreur
		if (EmlibFrameEdit.dataLineIndex == 0) then
			------------ nouvelle emote
			table.insert(EMLIB_SAVEDPERCHAR["emotes"], completeEmote(tblEmote, editBuiltin.checked))
		else
			------------ save existant
			local numLine = EmlibFrameEdit.dataLineIndex
			if (EMLIB_SAVEDPERCHAR["emotes"][numLine] ~= nil) then -- secu
				EMLIB_SAVEDPERCHAR["emotes"][numLine] = completeEmote(tblEmote, editBuiltin.checked)
			else
				errSaved = true
				Emlib.echoChatAppl("unexpected error #3900", false, "red")
			end
		end
		
		if (not errSaved) then
			Emlib.sortListAlpha()
			Emlib.scrollUpdate(EmlibUserScrollFrame, EMLIB_SAVEDPERCHAR["emotes"])
			Emlib.showHideEditUI(false)
		end
	else
		-------------------------------- erreur
		Emlib.echoChatAppl(Emlib.getLocal("Error in emote definition."), false, "red")
		_G["EmlibFrameEditZone_EditBox_"..errEmote]:SetFocus()
	end
end

function Emlib.btnEmoteDelete()
	local numLine = Emlib.strToNumber(EmlibFrameEdit.dataLineIndex)
	if (numLine > 0) then
		if (EMLIB_SAVEDPERCHAR["emotes"][numLine] ~= nil) then -- secu
			table.remove(EMLIB_SAVEDPERCHAR["emotes"], numLine)
			Emlib.scrollUpdate(EmlibUserScrollFrame, EMLIB_SAVEDPERCHAR["emotes"])
			Emlib.showHideEditUI(false)
		else
			errSaved = true
			Emlib.echoChatAppl("unexpected error #3950", false, "red")
		end
	else
		errSaved = true
		Emlib.echoChatAppl("unexpected error #3951", false, "red")
	end
end

---------------------------------------------------------------------------------------
------------------------------------ TOKEN EMOTES -------------------------------------
--[[
Emote Tokens for use with DoEmote, these are defined in FrameXML/ChatFrame.lua, as EMOTEx_TOKEN, where x is 1 to 170. They are not localised. 
EMOTE1_TOKEN = "AGREE";
EMOTE2_TOKEN = "AMAZE";
EMOTE3_TOKEN = "ANGRY";
--]]

function Emlib.createTokenScroll(parent, height) -- scroll liste token emotes
	frameTokenList = CreateFrame("Frame", "EmlibTokenScrollFrame", parent) -- définition scroll frame
	Emlib.setColoredFrame(frameTokenList, "333333", "222222", 0, 0.6)

	local scrollProps = { -- scroll frame properties (see _scroll.lua for full settings)
		-------------------- height: at least one of these properties must be set
		["height"] = height,
		--------------------
		["width"] = Emlib.getVar("tokenWidth"),
		["marginH"] = 1, -- px (margins from background frame : depends on template style)
		["marginV"] = 0, -- px (margins from background frame : depends on template style)
		["alwaysShowArrow"] = true, -- if true, always show scroller up and down arrows
		-------------------- lines
		["lineHeight"] = Emlib.getVar("vLine"), -- px
		["lineSpacing"] = 1, -- px
		-------------------- lines font
		["font"] = Emlib.getStuffPath("Myriad Condensed Web.ttf", true), -- "friz" | "arial" | pathToLocalFont accepted
		["fontSize"] = 11,
		["fontOutline"] = false,
		["fontColor"] = "CCC033",
		["fontAlign"] = "CENTER", -- "LEFT" | "RIGHT" | "CENTER"
		["fontAlpha"] = 0.7,
		-------------------- lines background
		["background"] = {0.1, 0.1, 0.1},
		["border"] = 0, -- only available if ["background"] is not a texture path
		["borderColor"] = nil, -- only available if ["background"] is not a texture path
		["alpha"] = 0.6,
		["mouseOverLineColor"] = "222222", -- visual color effect on mouseOver all the lines (nil for no mouseOver effect)
		-------------------- scroller buttons images
		["defaultScroller"] = false, -- set to true to standard WoW scroller
		["scrollerColor"] = "222233", -- unused if defaultScroller set to true
		["scrollerBarAlpha"] = 0.4, -- unused if defaultScroller set to true
	}
	
	local scrollColumns = { -- columns definition and properties (see _scroll.lua for full settings)
		-- column order, do not use name !
		[1] = { -- cmd
			---------- required
			["width"] = nil, -- toute la largeur
			["type"] = "Button", -- "Frame" (default), "Button" or "EditBox"
			---------- optional script (for button "OnClick" or EditBox "OnEnterPressed")
			["script"] = "Emlib.clickLineTokenList",
		},
	}
	
	Emlib.scrollCreate(frameTokenList, scrollProps, scrollColumns, nil) -- creates structure (empty scroll lines)
end

function Emlib.clickLineTokenList(scrollFrameName, clickedColumnName, mouse, lineIndex) -- do not modify args !
	-- mouse: "LeftButton" | "RightButton"
	-- if (IsModifierKeyDown()...)
	local scrollFrame = _G[scrollFrameName] -- scroll frame
	local clickedFrame = _G[clickedColumnName] -- .text and .texture available
	
	
	local dataLineIndex, dataLine = Emlib.scrollGetDataLine(scrollFrame, lineIndex)
	if (not Emlib.isEmpty(dataLine)) then
		Emlib.flash(clickedFrame.texture, 0.2, 0)
		if (IsShiftKeyDown()) then
			editCmd:SetText(editCmd:GetText()..dataLine)
			editCmd:SetFocus()
		elseif (IsAltKeyDown()) then
			editCmdTarget:SetText(editCmdTarget:GetText()..dataLine)
			editCmdTarget:SetFocus()
		else
			Emlib.playEmote(Emlib.getNewEmote(dataLine, dataLine, dataLine, 1, 0))
		end
	else
		Emlib.echoChatAppl("unexpected error #3800", false, "red")
	end
end

function Emlib.getAvailableTokenEmotes() -- renvoit une table des emotes localisées
	local tbl = {}
	for i = 1, 400 do -- limite = 171 ?
		local tmp = _G["EMOTE"..i.."_CMD1"]
		if (tmp == nil) then
			break
		else
			table.insert(tbl, tmp)
		end
	end
	table.sort(tbl)
	return tbl
end